#!/usr/bin/env python
# -*- coding: utf-8 -*-
# wirkungsModell.py
# © Claus Albowski
from export.dialoge import * # eigene

#### Klasse für Faktoren (Groessen) ====================================
class Faktor(object):
    '''Klasse für Faktoren (Groessen)'''

    # Klassenvariable fuer alle Faktornamen und Faktoren
    faktorNamen=[]
    faktoren=[]

    @staticmethod
    def GibAlleFaktorNamen():
        return Faktor.faktorNamen

    @staticmethod
    def GibFaktorZuNamen(name):
        for faktor in Faktor.faktoren:
            if name==faktor.GibNamen():
                return faktor

    @staticmethod
    def LoescheFaktor(name):
        '''loescht den Namen aus der Liste alleFaktoren'''
        try:
            faktor=Faktor.GibFaktorZuNamen(name)
            Wirkung.LoescheWirkungenZuFaktor(name)
            Faktor.faktoren.remove(faktor)
            Faktor.faktorNamen.remove(name)
        except ValueError as e:
            WirkungsModell.GibWirkungsModell().Meldung(str(e))
    
    @staticmethod
    def LoescheAlleFaktoren():
        alleNamen=[]+Faktor.GibAlleFaktorNamen() # echte Kopie!
        for name in alleNamen:
            Faktor.LoescheFaktor(name)

    @staticmethod
    def PruefeNamen(name):
        if name in Faktor.faktorNamen:
            Faktor.GibFaktorZuNamen(name).Meldung("Name existiert bereits!")
            return False
        if name in globals().keys(): return False
        try:
            try:
                auswertung=eval(name)
            except NameError as e:
                # alles ok, keine Zahl
                return True
        except SyntaxError as ee:
            WirkungsModell.GibWirkungsModell().Meldung(str(ee))
            return False
        if auswertung.__class__ in (int,float,complex):
            WirkungsModell.GibWirkungsModell().Meldung("Keine Zahl verwenden!")
            return False
        return True
    
    #### Konstruktor von Faktor
    def __init__(self, name):
        '''fuer die Wirkungsdiagramme werden die Faktoren
        und die Namen der Faktoren gespeichert'''
        while not self.PruefeNamen(name):
            name=input(name+' aendern in: ')
        self.__name=name
        self.__typ = None # Bestand, Fluss, Parameter oder Zwischengroesse
        self.__vonAufNamen = (None,None) # Fluesse haben von und auf
        Faktor.faktoren.append(self)
        Faktor.faktorNamen.append(name)

    def GibTyp(self):
        return self.__typ

    def SetzeTyp(self, typ, von=None, auf=None):
        if typ in ['Bestand', 'Parameter','Zwischengroesse']:
            self.__typ = typ
        elif typ =='Fluss':
            if (von!=None) and (auf!=None):
                self.__typ = typ
                self.__vonAufNamen = von, auf

    def GibVon(self):
        return self.__vonAufNamen[0]

    def GibAuf(self):
        return self.__vonAufNamen[1]

    def GibNamen(self):
        try:
            return self.__name
        except AttributeError as e:
            self.Meldung("Name ist nicht richtig initialisiert:\n"
                         +"SetzeNamen() mit String notwendig!")
    
    def SetzeNeuenNamen(self, neuerName):
        if not(self.PruefeNamen(neuerName)): return False
        alterName=self.GibNamen()
        Faktor.faktorNamen.remove(alterName)
        self.__name=neuerName
        Faktor.faktorNamen.append(neuerName)
        return True

    def Meldung(self, text):
        WirkungsModell.GibWirkungsModell().Meldung(text)

#### Klasse fuer Wirkungen -------------------------------------------
class Wirkung():
    '''Klasse für Wirkungen'''
    
    wirkungen=[]

    @staticmethod
    def HatWirkung(vonName, aufName):
        '''Gibt die Wirkung zurueck, falls sie existiert, sonst False'''
        for wirkung in Wirkung.wirkungen:
            if ((wirkung.GibVon()==vonName)
                and (wirkung.GibAuf()==aufName)):
                return wirkung
        return False
    
    @staticmethod
    def GibAlleWirkungen():
        return Wirkung.wirkungen
    
    @staticmethod
    def LoescheWirkung(vonName, aufName):
        wirkung=Wirkung.HatWirkung(vonName, aufName)
        if wirkung:
            Wirkung.wirkungen.remove(wirkung)

    @staticmethod
    def LoescheAlleWirkungen():
        wirkungen=[]+Wirkung.GibAlleWirkungen() # echte Kopie
        for wirkung in wirkungen:
            Wirkung.wirkungen.remove(wirkung)
    
    @staticmethod
    def LoescheWirkungenZuFaktor(faktorName):
        '''Faktoren, die entfernt werden sollen, dürfen keine Wirkungen haben.'''
        for wirkung in Wirkung.wirkungen:
            if wirkung.GibVon()==faktorName:
                Wirkung.wirkungen.remove(wirkung)
            if wirkung.GibAuf()==faktorName:
                Wirkung.wirkungen.remove(wirkung)

    #### Konstruktor fuer Wirkungen ------------------------------------    
    def __init__(self, vonName, aufName, art='+'):
        '''Eine Wirkung muss so initialisiert werden, dass sowohl der
        Ausgangs-Faktor als auch der Ziel-Faktor gegeben sind.'''
        self.undefiniert = False
        self.istFluss = False # fuer den Export wichtig
        if Wirkung.HatWirkung(vonName, aufName):
            self.Meldung('Wirkung existiert bereits')
            self.undefiniert = True
        if not(vonName in Faktor.GibAlleFaktorNamen()):
            self.Meldung(vonName+' existiert nicht')
            self.undefiniert = True
        if not(aufName in Faktor.GibAlleFaktorNamen()):
            self.Meldung(aufName+' existiert nicht')
            self.undefiniert = True
        else:
            self.__von=vonName
            self.__auf=aufName
            self.__art=art
            Wirkung.wirkungen.append(self)

    def IstFluss(self, neuerWert=False):
        if neuerWert: self.istFluss = True
        return self.istFluss

    def GibVon(self):
        return self.__von

    def GibAuf(self):
        return self.__auf

    def GibArt(self):
        return self.__art

    def AendereArt(self):
        if self.__art=='+': self.__art='-'
        else: self.__art='+'

    def ZeigeWirkung(self):
        return self.GibVon()+' --> '+self.GibAuf()+' '+self.GibArt()

    def FaktorNameGeaendert(self, alterName, neuerName):
        if self.GibVon()==alterName:
            self.__von=neuerName
        elif self.GibAuf()==alterName:
            self.__auf=neuerName

    def Meldung(self, text):
        WirkungsModell.GibWirkungsModell().Meldung(text)

#### WirkungsModell ----------------------------------------------------------------
class WirkungsModell():
    '''WirkungsModell erstellt das Modell fuer ein Wirkungsdiagramm.'''

    dasModell = None

    @staticmethod
    def GibWirkungsModell(view=None):
        '''Singleton ermoeglichen'''
        if WirkungsModell.dasModell == None:
            WirkungsModell.dasModell = WirkungsModell()
            WirkungsModell.dasModell.SetzeView(view)
        return WirkungsModell.dasModell

    def __init__(self):
        '''Modell fuer ein Wirkungsdiagramm
        Die Faktoren und Wirkungen werden in den Klassen gehalten'''
        self.__modellName=None
        self.__modellDaten=[] # fuer die Ausgabe

    def SetzeTyp(self, name, typ, von=None, auf=None):
        if not (typ in ['Bestand', 'Fluss', 'Parameter','Zwischengroesse']): return
        if name in Faktor.GibAlleFaktorNamen():
            faktor=Faktor.GibFaktorZuNamen(name)
            faktor.SetzeTyp(typ,von,auf)
            if typ=='Fluss':
                hat=Wirkung.HatWirkung(faktor.GibNamen(),auf)
                if hat: hat.IstFluss(True)
                hat=Wirkung.HatWirkung(faktor.GibNamen(),von)
                if hat: hat.IstFluss(True)

    def GibFaktorenZuTyp(self,typ):
        faktoren=[]
        for name in Faktor.GibAlleFaktorNamen():
            faktor=Faktor.GibFaktorZuNamen(name)
            if faktor.GibTyp()==typ:
                faktoren.append(name)
        return faktoren

    def SetzeModellNamen(self, name):
        assert name.__class__==str, "String eingeben!"
        self.__modellName=name

    def GibModellNamen(self):
        return self.__modellName

    def AendereFaktorNamen(self, alterName, neuerName):
        faktor=Faktor.GibFaktorZuNamen(alterName)
        if faktor.SetzeNeuenNamen(neuerName): ## Wirkungen informieren:
            for wirkung in Wirkung.GibAlleWirkungen():
                wirkung.FaktorNameGeaendert(alterName,neuerName)
            return True
        return False

    def GibAlleFaktorNamen(self):
        return Faktor.GibAlleFaktorNamen()

    def NeuerNameOK(self, name):
        if Faktor.PruefeNamen(name):
            return True
        return False

    def NeuerFaktor(self, name):
        '''Alle Faktoren sind in Faktor gespeichert'''
        if self.NeuerNameOK(name):
            return Faktor(name)
        return None
        
    def NeueWirkung(self, vonName, aufName, art='+'):
        '''Wirkungen werden in der Klasse gehalten'''
        if vonName==aufName: return None
        wirkung=Wirkung(vonName, aufName, art)
        if wirkung.undefiniert: return None
        return wirkung

    def LoescheFaktor(self,name):
        '''entfernt das Faktorobjekt zum Namen
        und entfernt die zugehoerigen wirkungen'''
        faktor=Faktor.GibFaktorZuNamen(name)
        if faktor!=None:
            ## Alle Wirkungen loeschen, die diesen Faktor betreffen.
            for wirkung in Wirkung.GibAlleWirkungen():
                if (name==wirkung.GibVon()) or (name==wirkung.GibAuf()):
                    self.LoescheWirkung(wirkung.GibVon(),wirkung.GibAuf())
            Faktor.LoescheFaktor(name)
        else:
            self.Meldung("Faktor "+name+" existiert nicht!")
        
    def LoescheWirkung(self, vonName, aufName):
        wirkung=Wirkung.HatWirkung(vonName, aufName)
        if not(wirkung): return False # existiert nicht
        Wirkung.LoescheWirkung(vonName, aufName)

    def LoescheModell(self):
        Wirkung.LoescheAlleWirkungen()
        Faktor.LoescheAlleFaktoren()

#### Abschnitt für die Modelldefinition -----------------------------
    def GibModellDaten(self):
        return self.__modellDaten

    def ZeigeModellDaten(self):
        '''Modelldaten zeilenweise lesbar'''
        alleDaten=self.__modellDaten
        zeilen=[]
        for daten in alleDaten:
            zeile=''
            for wort in  daten:
                zeile+=str(wort)+' '
            zeilen.append(zeile)
        return zeilen
    
    
    def SichereModellDaten(self):
        self.__modellDaten=[]
        if self.__modellName==None:
            self.SetzeModellNamen('Neu')
        for faktor in Faktor.faktoren:
            if faktor.GibTyp()=='Bestand':
                ## der Anfangswert von Bestaenden wird auf 0 gesetzt
                faktorDaten=('Bestand', faktor.GibNamen(), 0.0)
                ## erst alle Faktoren definieren
                self.__modellDaten.append(faktorDaten)
        for faktor in Faktor.faktoren:
            if faktor.GibTyp()=='Fluss':
                faktorDaten=('Fluss', faktor.GibNamen(),
                             faktor.GibVon(),
                             faktor.GibAuf(),
                             '')
                self.__modellDaten.append(faktorDaten)
            elif faktor.GibTyp()=='Parameter':
                ## der Wert von Parametern wird auf 0 gesetzt
                faktorDaten=('Parameter', faktor.GibNamen(), 0.0,)
                self.__modellDaten.append(faktorDaten)
            elif faktor.GibTyp()=='Zwischengroesse':
                ## Zwischengroessen haben keine Werte, Terme werden gesondert gesetzt
                faktorDaten=('Zwischengroesse', faktor.GibNamen())
                self.__modellDaten.append(faktorDaten)
            elif faktor.GibTyp()==None:  # undefinierter Faktor
                faktorDaten=('Faktor',faktor.GibNamen())
                self.__modellDaten.append(faktorDaten)

        for wirkung in Wirkung.GibAlleWirkungen():
            if not(wirkung.IstFluss()):
                self.__modellDaten.append(
                    ('Wirkung',wirkung.GibVon(),wirkung.GibAuf()))
        
    def ErzeugeModellAusDaten(self, modell_Daten):
        for daten in modell_Daten:
            if daten[0]=='Faktor':
                self.NeuerFaktor(daten[1])
            elif daten[0]=='Wirkung':
                self.NeueWirkung(daten[1],daten[2],daten[3])
            else:
                self.Meldung('Fehler bei Modelldefinition: '+daten[0])

    def ModellNameAusDateiNamen(self, dateiname):
        # Pfad abspalten
        if '\\' in dateiname: # windows-Pfadangabe
            liste=dateiname.rsplit('\\')
        else:
            liste=dateiname.rsplit('/')
        liste.reverse()
        name=liste[0]
        if '.' in name:
            name=name.rsplit('.')[0]
        return name

#### Abschnitt für das Laden und Speichern von Modelldaten -----------------
    def Speichere(self, dateiname='', modellDaten=[]):
        self.SichereModellDaten()
        if dateiname=='':
            if self.__modellName==None: return
            dateiname=self.__modellName+'.wirkdia'
        try:
            f=open(dateiname,'w')
        except IOError as e:
            self.Meldung('Datei existiert nicht, keine Daten geschrieben')
            return None
        if modellDaten!=[]: f.write(str(modellDaten))
        else: f.write(str(self.__modellDaten))
        f.close()

    def Lies(self, dateiname):
        try:
            f=open(dateiname,'r')
        except IOError as e:
            self.Meldung('Datei existiert nicht, keine Daten gelesen')
            return None
        modell_Daten_String=f.read()
        f.close()
        self.__modellName=self.ModellNameAusDateiNamen(dateiname)
        self.__modell_Daten=eval(modell_Daten_String)
        self.ErzeugeModellAusDaten(self.__modell_Daten)
        return self.__modell_Daten

#### Zusaetzliches -----------------------------
    def ZeigeHilfe(self, n=1):
        if n==1:
            hinweise=("TODO"
                      )
        elif n==2:
            hinweise=("TODO.")
        return hinweise
    
    def SetzeView(self, view):
        self.__view=view
        
    def Meldung(self, text):
        if self.__view==None: print(text)
        else:
            MeldungDialog(self.__view,text)
            

#### Tests
if __name__=='__main__':
##    w=WirkungsModell.GibWirkungsModell()
##    w.NeuerFaktor('Abkuehlung')
##    w.NeuerFaktor('Temperatur')
##    w.NeuerFaktor('Umgebungstemperatur')
##    w.NeueWirkung('Abkuehlung','Temperatur')
##    w.NeueWirkung('Temperatur','Abkuehlung')
##    w.NeueWirkung('Umgebungstemperatur','Abkuehlung')
##    w.SichereModellDaten()
##    print(w.ZeigeModellDaten())
##    print()
##    w.SetzeTyp('Temperatur', 'Bestand')
##    w.SetzeTyp('Abkuehlung', 'Fluss','Temperatur', 'Senke')
##    w.SetzeTyp('Umgebungstemperatur', 'Parameter')
##    w.SichereModellDaten()
##    print(w.ZeigeModellDaten())
##    print(w.GibFaktorenZuTyp('Bestand'))
##    print(w.GibFaktorenZuTyp('Fluss'))
##    print(w.GibFaktorenZuTyp('Parameter'))

    w=WirkungsModell.GibWirkungsModell()
    w.NeuerFaktor('Strecke')
    w.NeuerFaktor('Geschwindigkeit')
    w.NeuerFaktor('Beschleunigung')
    w.NeueWirkung('Geschwindigkeit','Strecke')
    w.NeueWirkung('Beschleunigung','Geschwindigkeit')
    w.SichereModellDaten()
    print(w.ZeigeModellDaten())
    w.SetzeTyp('Strecke', 'Bestand')
    w.LoescheWirkung('Geschwindigkeit','Strecke')
    w.NeuerFaktor('Aenderung_Strecke')
    w.NeueWirkung('Geschwindigkeit','Aenderung_Strecke')
    w.NeueWirkung('Aenderung_Strecke','Strecke')
    w.SetzeTyp('Aenderung_Strecke', 'Fluss', 'Quelle', 'Strecke')

    w.SetzeTyp('Geschwindigkeit', 'Bestand')
    w.SetzeTyp('Beschleunigung', 'Parameter')
    w.LoescheWirkung('Beschleunigung','Geschwindigkeit')
    w.NeuerFaktor('Aenderung_Geschwindigkeit')
    w.NeueWirkung('Beschleunigung','Aenderung_Geschwindigkeit')
    w.NeueWirkung('Aenderung_Geschwindigkeit','Geschwindigkeit')
    w.SetzeTyp('Aenderung_Geschwindigkeit', 'Fluss', 'Quelle','Geschwindigkeit')
    print(Wirkung.HatWirkung('Aenderung_Geschwindigkeit','Geschwindigkeit').IstFluss())
    w.SichereModellDaten()
    print("\n")
    print(w.GibModellDaten())
    print("\n")
    for zeile in w.ZeigeModellDaten():
        print(zeile)
##    print('Bestaende: '+str(w.GibFaktorenZuTyp('Bestand')))
##    print('Fluesse: '+str(w.GibFaktorenZuTyp('Fluss')))
##    print('Parameter:'+str(w.GibFaktorenZuTyp('Parameter')))
##    print('Zwischengroessen: '+str(w.GibFaktorenZuTyp('Zwischengroesse')))


